import argparse
import numpy as np
import torch
from hide_in_plain_sight import hide_in_plain_sight
from deidentifier_model import deidentifier_model
from multiprocessing import Process
import time
import random
import os
from connect_input_utils import parse_input
from connect_output_utils import parse_output
import pandas as pd
import logging
import multiprocessing
import sys
import shutil


def parse_args():
    parser = argparse.ArgumentParser(description="Deidentify radiology reports")
    parser.add_argument(
        "--device_list",
        nargs="+",
        default=["cuda:0"],
        help="Devices to run the transformer(s) model on, can provide several spaced device names to scale. Must be one of cpu, mps, cuda, or cuda:device_number",
        # required=True,
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        default=0,
        help="Number of workers for the transformer",
    )
    parser.add_argument(
        "--num_cpu_processes",
        type=int,
        default=1,
        help="If running on cpu, can split the deidentification between several processes",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=32,
        help="Batch size for inference with the transformer model",
    )
    parser.add_argument(
        "--max_number_of_reports_to_be_processed_in_parallel",
        type=int,
        default=50000,
        help="Max number of reports to process in parallel, should be lowered if "
        + "encountering memory issues, can be augmented if lots of remaining memory",
    )
    # parser.add_argument(
    #     "--input_file_path",
    #     type=str,
    #     default="/scratch/reports.npy",
    #     help="Path to input file, must be .npy",
    #     # required=True,
    # )
    # parser.add_argument(
    #     "--output_file_path",
    #     type=str,
    #     default="/scratch/deidentified_reports.npy",
    #     help="Path to output file, must be .npy",
    #     # required=True,
    # )
    parser.add_argument(
        "--hospital_list",
        nargs="+",
        default=['SiteA', 'SiteB', 'SiteC'],
        help="Hospitals to be parsed based on lower-case matching",
        required=False,
    )
    parser.add_argument(
        "--vendor_list",
        nargs="+",
        help="Vendors to be parsed based on lower-case matching",
        required=False,
    )
    ##Connect args
    parser.add_argument(
        "--input_dir",
        type=str,
        default='/input',
        help="Connect input directory path",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default='/output',
        help="Connect output directory path",
    )
    parser.add_argument(
        "--scratch_dir",
        type=str,
        default='/scratch',
        help="Connect scratch directory path",
    )
    ##Connect args
    parser.add_argument(
        "--input_filepath",
        type=str,
        help="full path to Connect input file",
    )

    return parser.parse_args()


def check_args_validity(args):
    assert args.batch_size >= 0
    # assert args.input_file_path[-4:] == ".npy"
    # assert args.output_file_path[-4:] == ".npy"
    assert len(args.device_list) > 0
    for device in args.device_list:
        if device not in ["cpu", "mps", "cuda"]:
            assert device[:5] == "cuda:"
            assert device[5:].isnumeric()


def deidentifier_model_and_hide_in_plain_sight(
    ref_dir, file_seed, device, num_workers, batch_size, hospital_list, vendor_list, logger
):
    try:
        deidentifier_model(
            file_seed, device, num_workers, batch_size, hospital_list, vendor_list, logger
        )
        hide_in_plain_sight(ref_dir, file_seed)
    except Exception as e:
        logger.error(f'Error running de-id algorithm: {e}')


def generate_output_files(file_seed_list, output_file_path):
    reports = []
    labeled_reports = []
    deidentified_reports = []
    phi_lengths = []

    for file_seed in file_seed_list:
        with open("original_reports" + file_seed + ".npy", "rb") as f:
            reports.append(np.load(f, allow_pickle=True))
            #print('OG: ' , reports[0][0])
        os.remove("original_reports" + file_seed + ".npy")

        with open("labeled_reports" + file_seed + ".npy", "rb") as f:
            labeled_reports.append(np.load(f, allow_pickle=True))
            #print('LABELED: ' , labeled_reports[0][0])
        os.remove("labeled_reports" + file_seed + ".npy")

        with open("deidentified_reports" + file_seed + ".npy", "rb") as f:
            deidentified_reports.append(np.load(f, allow_pickle=True))
            #print('DEID: ' , deidentified_reports[0][0])
        os.remove("deidentified_reports" + file_seed + ".npy")

        with open("phi_lengths" + file_seed + ".npy", "rb") as f:
            phi_lengths.append(np.load(f, allow_pickle=True))
        os.remove("phi_lengths" + file_seed + ".npy")

    reports = np.concatenate(reports)
    labeled_reports = np.concatenate(labeled_reports)
    deidentified_reports = np.concatenate(deidentified_reports)
    phi_lengths = np.concatenate(phi_lengths)

    with open(output_file_path, "wb") as f:
        np.save(f, deidentified_reports, allow_pickle=True)

    print("finished writing the deidentified reports")

    with open("deidentification_details_for_review_reports.npy", "wb") as f:
        np.save(f, reports, allow_pickle=True)

    with open("deidentification_details_for_review_labeled_reports.npy", "wb") as f:
        np.save(f, labeled_reports, allow_pickle=True)

    with open("deidentification_details_for_review_phi_lengths.npy", "wb") as f:
        np.save(f, phi_lengths, allow_pickle=True)

    print("Only task remaining is saving a .csv file for review purposes")
    print("This step can take a long time, you can interrupt if needed")
    print(
        "All the data of the .csv file being written has already been saved in separate .npy files"
    )

    df_for_review = pd.DataFrame(
        [reports, labeled_reports, deidentified_reports, phi_lengths]
    ).transpose()
    df_for_review.rename(
        columns={
            0: "original_reports",
            1: "labeled_reports",
            2: "deidentified_reports",
            3: "phi_lengths",
        },
        inplace=True,
    )
    df_for_review.to_csv("deidentification_details_for_review.csv", index=False)


def main(args):
    #determine os
    #for python exe 'app' needs to be appended to paths for reading models and txt/csv files
    try:
        ref_dir = os.path.join(sys._MEIPASS, 'app')
        print(f'reference directory is {ref_dir}')
    except:
        ref_dir = ''
    start = time.time()
    os.environ['TRANSFORMERS_CACHE'] = os.path.join(ref_dir, 'model_files')
    print('transformers cache: ' , os.environ['TRANSFORMERS_CACHE'])
    print('ref dir: ', ref_dir)
    #input_dir = args.input_dir
    scratch_dir = args.scratch_dir
    output_dir = args.output_dir
    input_filepath = args.input_filepath

    log_filepath = os.path.join(output_dir, 'standford_report_deid' + '_' + str(start) + '.log')
    logger = logging.getLogger('standford_report_deid')
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)-8s %(message)s', datefmt='%a, %d %b %Y %H:%M:%S', filename=log_filepath, filemode='w')
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)-8s %(message)s', datefmt='%a, %d %b %Y %H:%M:%S', filename=log_filepath, filemode='w')
    
    logger.info(f'Starting at {start}')
    
    try:
        #parse connect input
        logger.info('Reformatting Assess AI input files')
        parse_input(input_filepath, scratch_dir, logger)
        
        #with open(args.input_file_path, "rb") as f:
        with open(os.path.join(scratch_dir, 'reports.npy'), "rb") as f:
            reports = np.load(f, allow_pickle=True)
    
        logger.info('Input files reformatted. Begining the deidentification process.')
        logger.info(f"Running inference on {len(reports)} reports.")
            
        device_list = (
            args.device_list
            if (len(args.device_list) > 1 or args.device_list[0] != "cpu")
            else ["cpu" for _ in range(args.num_cpu_processes)]
        )
    
        device_list = [torch.device(device) for device in device_list]
        file_seed_list = []
        number_batches = (
            len(reports) // args.max_number_of_reports_to_be_processed_in_parallel + 1
        )
    
        for batch_index in range(number_batches):
            file_seed_list_local = []
    
            print(
                "Processed so far",
                args.max_number_of_reports_to_be_processed_in_parallel * batch_index,
                "reports",
            )
    
            number_reports_in_the_batch = (
                args.max_number_of_reports_to_be_processed_in_parallel
                if batch_index < number_batches - 1
                else (
                    len(reports)
                    - args.max_number_of_reports_to_be_processed_in_parallel * batch_index
                )
            )
    
            if number_reports_in_the_batch == 0:
                assert batch_index == number_batches - 1
                assert args.max_number_of_reports_to_be_processed_in_parallel == len(
                    reports
                )
                continue
    
            number_reports_per_file = number_reports_in_the_batch // len(device_list)
    
            # Prepare files to be processed in each separate process
            for i, device in enumerate(device_list):
                file_seed = (
                    device.type
                    + "device"
                    + str(i)
                    + "device"
                    + str(batch_index)
                    + "batch_index"
                    + str(random.randint(2394492340, 23944923402394492340))
                )
    
                if i + 1 < len(device_list) and number_reports_per_file == 0:
                    continue
    
                file_seed_list.append(file_seed)
                file_seed_list_local.append(file_seed)
    
                with open(
                    "original_reports" + file_seed + ".npy",
                    "wb",
                ) as f:
                    np.save(
                        f,
                        np.array(
                            reports[
                                args.max_number_of_reports_to_be_processed_in_parallel
                                * batch_index
                                + i
                                * number_reports_per_file : (
                                    (
                                        args.max_number_of_reports_to_be_processed_in_parallel
                                        * batch_index
                                        + (i + 1) * number_reports_per_file
                                    )
                                    if i + 1 < len(device_list)
                                    else (
                                        args.max_number_of_reports_to_be_processed_in_parallel
                                        * batch_index
                                        + number_reports_in_the_batch
                                    )
                                )
                            ]
                        ).astype("object"),
                        allow_pickle=True,
                    )
    
            processes = []
            
            for file_seed, device in zip(file_seed_list_local, device_list):
                p = Process(
                    target=deidentifier_model_and_hide_in_plain_sight,
                    args=(
                        ref_dir,
                        file_seed,
                        device,
                        args.num_workers,
                        args.batch_size,
                        args.hospital_list if args.hospital_list is not None else [],
                        args.vendor_list if args.vendor_list is not None else [],
                        logger
                    ),
                )
                p.start()
                processes.append(p)
                # Running without parallelization
                # deidentifier_model_and_hide_in_plain_sight(
                # file_seed,
                # device,
                # args.num_workers,
                # args.batch_size,
                # args.hospital_list,
                # args.vendor_list,
                # )
                # pass
    
            for p in processes:
                p.join()
    
            time.sleep(30)  # giving time for the processes to be cleaned
            logger.info('Inference completed. Reformatting algorithm output.')
    
        #generate_output_files(file_seed_list, args.output_file_path)
        generate_output_files(file_seed_list, os.path.join(scratch_dir, 'deidentified_reports.npy'))
        #copy algorithm output csv to connect output dir
        cwd = os.getcwd()
        shutil.copyfile(os.path.join(cwd,'deidentification_details_for_review.csv'), os.path.join(output_dir, 'deidentification_details_for_review.csv'))
        #add additional input args if needed
        parse_output(output_dir, input_filepath, os.path.join(scratch_dir, 'deidentified_reports.npy'), logger)
    
        logger.info(f"Ended execution, took {time.time() - start} seconds")

    except Exception as e:
        logging.error(e)
        #raise


if __name__ == "__main__":
    multiprocessing.freeze_support()
    args = parse_args()
    check_args_validity(args)
    main(args)
